import numpy as np
from scipy.integrate import solve_ivp
from scipy.optimize import curve_fit, least_squares
from scipy.stats import linregress
import matplotlib.pyplot as plt

#Checking for linearity
#Actually this won't work and I'm not dealing with it right now
def isLinear(specs, orders):
	return False
	if orders is None:
		return len(specs) == 1
	else:
		nonzeros, = np.where(np.array(orders) != 0)
		return len(nonzeros) == 1 and nonzeros[0] == 1

#Mainly a supporting class for Reaction; keep tracks of an individual rate in a reaction
#A Rate has a list of reactant/product indices but not their names
#Each reactant can be listed N times, which is easiest for elementary reactions
#But, it was written to be flexible - it can also take arbitrary orders for each reactant
#If N equivalents 
class Rate:

	#k is the rate constant for the reaction step modeled here
	#reactants gives the indices of reactants
	#products gives the indices of products
	#	Stoichiometry is indicated by repeats, i.e. if two equivalents of reactant/product
	#		is consumed/formed, the index needs to be listed twice
	#orders gives the order for each reactant, assumed to be 1 if non are given
	#	Not needed for elementary reactions
	#mutable is if the rate can be change when optimizing a Reaction intances's Rate laws
	def __init__(self, k, reactants, products, orders = None, mutable=True):
		self.k = k
		self.reactants = reactants
		if orders is not None:
			self.orders = orders
		else:
			self.orders = [1]*len(reactants)
		self.products = products
		self.linear = len(self.reactants) == 1 and len(self.products) == 1 and orders[0] == 1 #Too restrictive but I'll fix it later
		self.mutable = mutable
	
	#Setters and getters for rate constant k and mutable
	def setRate(self, k):
		self.k = k
	def getRate(self):
		return self.k
	def setMutable(self, mutable):
		self.mutable = mutable
	def getMutable(self):
		return self.mutable
	
	#Calculates the rate, given an array of concentrations
	#Remember an instance of Rate knows reactants/products by indices
	#returns the contribution of of this Rate to each species individual rate
	#	as an array of the same length of the given concentrations
	#(The contribution is 0 if an index is not in reactants or products)
	def calcRate(self, concs):
		deriv = self.k
		for i in range(len(self.reactants)):
			if concs[self.reactants[i]] <= 0:
				deriv = 0
				#print("Hit zero with concs", concs)
				break
			deriv *= concs[self.reactants[i]]**self.orders[i]
		toReturn = np.zeros(len(concs))
		for i in range(len(concs)):
			if i in self.reactants:
				toReturn[i] = -deriv*self.reactants.count(i)
			if i in self.products:
				toReturn[i] = +deriv*self.products.count(i)
		#print("\t", self.k, deriv, self.reactants, self.orders, self.products)
		return toReturn
	
	#Returns a format string template for giving the rate law for this rate
	#An instance of reaction can use format to insert names in place of indices	
	def rateLaw(self):
		print(self.reactants, self.products, self.orders)
		toReturn = ""
		bins = np.bincount(self.reactants)
		for i in range(len(bins)):
			if bins[i] != 0:
				totalOrder = 0
				for j in range(len(self.reactants)):
					if self.reactants[j] == i:
						totalOrder += self.orders[j]
				if totalOrder/1 != totalOrder//1:
					toReturn += "[{{{}}}]^{:.2f} ".format(i, totalOrder)
				else:
					toReturn += "[{{{}}}]^{} ".format(i, totalOrder)
		return toReturn[:-1]
	
	#Returns a format string template for listing the reactants for this rate
	#An instance of reaction can use format to insert names in place of indices
	def reactantStr(self):
		toReturn = ""
		bins = np.bincount(self.reactants)
		for i in range(len(bins)):
			if bins[i] != 0:
				toReturn += "{}[{{{}}}] + ".format(bins[i], i)
		return toReturn[:-3]

	#Returns a format string template for listing the products for this rate
	#An instance of reaction can use format to insert names in place of indices
	def productStr(self):
		toReturn = ""
		bins = np.bincount(self.products)
		for i in range(len(bins)):
			if bins[i] != 0:
				toReturn += "{}[{{{}}}] + ".format(bins[i], i)
		return toReturn[:-3]
	
#In theory leveraging linearity for efficiently would be nice
#But I haven't been able to quite develop that yet
class Linear(Rate):
	def __init__(self, k, reactant, products, spectators = None):
		self.k = k
		self.reactant = reactant
		self.products = products
		self.spectators = spectators if spectators is not None else []
	def calcRate(self, concs):
		toReturn = np.zeros(len(concs))
		rate = self.k*concs[self.reactant]
		toReturn[self.reactant] -= rate
		for spectator in self.spectators:
			toReturn[spectator] -= rate
		for product in self.products:
			toReturn[product] += rate
		return toReturn

#Keeps track of the Rate reversible reaction step
#All is similar to, except:
#	It needs a forward rate and an equilibrium constant
#	You now have orders for prodcuts too
#Keeps track of forward/backward reactions as two seperrate Rate instances
class Reversible(Rate):
	def __init__(self, k, K, reactants, products, reactantOrders = None, productOrders = None, mutable=True):
		self.forward = Rate(k, reactants, products, reactantOrders, mutable)
		self.backward = Rate(k/K, products, reactants, productOrders, mutable)
		self.K = K
	def setMutable(self, mutable):
		self.forward.setMutable(True)
		self.backward.setMutable(True)
	def getMutable(self):
		return self.forward.getMutable()
	def calcRate(self, concs):
		return self.forward.calcRate(concs) + self.backward.calcRate(concs)
	def setRate(self, k):
		self.forward.k = k
		self.backward.k = k/self.K
	def getRate(self):
		return self.forward.k
	def rateLaw(self): return self.forward.rateLaw(), self.backward.rateLaw()
	def reactantStr(self): return self.forward.reactantStr()
	def productStr(self): return self.forward.productStr()

#Again, would be nice but isn't developed right yet
class LinearReversible(Linear, Reversible):
	def __init__(self, k, K, reactant, product):
		self.k = k
		self.K = K
		self.reactant = reactant
		self.product = product
		self.linear = True
	def calcRate(self, concs):
		forward = self.k*concs[self.reactant]
		backward = self.k/self.K*concs[self.product]
		toReturn = np.zeros(len(concs))
		toReturn[self.reactant] = -forward + backward
		toReturn[self.product] = forward - backward
		return toReturn
	def asMatrix(self, dim):
		toReturn = np.zeros((dim,dim))
		toReturn[self.reactant, self.reactant] = -self.k
		toReturn[self.product, self.reactant] = self.k
		toReturn[self.reactant, self.product] = self.k/self.K
		toReturn[self.product, self.product] = -self.k/self.K
		return toReturn

#So apparently I left no comments when I first wrote this and I think a few things need explaining
#This class is meant to handle both the loading of and the simulation of a variety of rates
#Rates keep track of reactants/products only as numerical indices; reactants have the mapping from string name to numeric index (species)
#It also has the simulation functions, so multiple rates can easily be taken account of when doing simulations
#It looks like you do need to give C0 as an array; I should change that at some point
class Reaction:
	
	def __init__(self):
		
		#Keeping track of rate laws
		self.rates = []
		
		#Keeping track of species in the reaction
		self.species = {}
		self.C0 = {}
		self.concentrations = None
		
		#Keeping track of time
		self.t_init = 0
		self.t_eval = None
		self.t_final = np.inf
		
		#Keeping track of fit parameters
		self.sol = None
		self.k_fit = None
		self.C0_fit = None
		self.t_fit = False
		self.cov = None
	
	#Prunes the list of all rates to give just the ones that can have rate constants optimized
	def mutableRates(self):	
		toReturn = []
		for rate in self.rates:
			if rate.getMutable():
				toReturn.append(rate)
		return toReturn
	
	#Parses a Rate from a string, using newRate to add it to the Reaction
	#Format to give an elementary irreversible reaction:
	#   A B -k-> C
	#Format to give an elementary reversible reaction:
	#  A B <-k/K-> C
	#In any case, species are separated by spaces and the rate/equilibrium
	#	constants are inside a "2-dash arrow"
	#To add stoichiometry/orders, you use brackets:
	#  "2{A}1" would give 2 of A reacting with order 1
	def parseRate(self, str, fixed=False):
		
		#Extract the parts of str corresponding to reactants, the reaction arrow, and the products
		pieces = str.split(">")
		for j in range(2,len(pieces[0])+2):
			if j == len(pieces[0])+1 or pieces[0][-j] == " ":
				reversible = (pieces[0][-j+1] == "<")
				reacting = pieces[0][:-j].strip().split()
				arrow = pieces[0][-j+(3 if reversible else 2):-1]
				produced = pieces[1].strip().split()
				#print("Reactants:", reacting)
				#print("Arrow:", arrow)
				#print("Products:", produced)
				break
		
		#Now parse each of the reactants and products in reacting and produced
		reactants = []
		reactantOrders = []
		products = []
		productOrders = []
		def parseSpec(spec):
			stoich = 1
			order = 1
			if "{" in spec:
				i = spec.index("{")
				j = spec.index("}")
				if i > 0: stoich = int(spec[:i])
				if j+1 < len(spec): order = float(spec[j+1:])
				spec = spec[i+1:j]
			else:
				spec = spec
			return spec, stoich, order
		for reactant in reacting:
			spec, stoich, order = parseSpec(reactant)
			reactants += [spec]*stoich
			if stoich == order:
				reactantOrders += [1]*stoich
			else:
				reactantOrders += [order/stoich]*stoich
		for product in produced:
			spec, stoich, order = parseSpec(product)
			products += [spec]*stoich
			if stoich == order:
				productOrders += [1]*stoich
			else:
				productOrders += [order/stoich]*stoich
		
		#Finally get k (and K if reversible) and add newRate to self
		if reversible:
			arrow = arrow.split("/")
			self.newRate(float(arrow[0]), reactants, products, reactantOrders, productOrders, float(arrow[1]), fixed=fixed)
		else:
			self.newRate(float(arrow), reactants, products, reactantOrders, fixed=fixed)
	
	#Adds a rate to the reaction, generating numeric indices for new reactants/products encountered
	#k is the forward rate constant
	#reactants are the names of reactants, their indices will either be retrieved or generated
	#products are the names of products; their indiced will be either retrieved or generated
	#reactantOrders gives the orders of reactants (1 if not given)
	#productOrders gives the orders of products, if reversible (1 if not given)
	#reversible signals a reversible reaction or not:
	#	None if irreversible
	#	The rate constant if it is reversible
	#	fixed gives if the rate is not to be varied during optimization
	def newRate(self, k, reactants, products, reactantOrders=None, productOrders=None, reversible=None, fixed=False):
		
		#Convert the string names of reactants and products to numeric indicies
		reactantIndices = []
		productIndices = []
		for reactant in reactants:
			self.addSpecies(reactant)
			reactantIndices.append(self.species[reactant])
		for product in products:
			self.addSpecies(product)
			productIndices.append(self.species[product])
		
		#Check for linearity, for now it is always false cause this feature isn't up to scratch yet
		linear = isLinear(reactantIndices, reactantOrders) and (reversible is None or isLinear(productIndices, productOrders))
		
		#Return the rate, be it reversible or not
		if reversible is None:
			#if linear:
			#	pass #You'll need to separate out spectators at some point
			toAppend = Rate(k, reactantIndices, productIndices, reactantOrders, not fixed)
		else:
			#if linear:
			#	pass #Again I need to do this when I actually make  isLinear work
			toAppend = Reversible(k, reversible, reactantIndices, productIndices, reactantOrders, productOrders, not fixed)
		self.rates.append(toAppend)
	
	#Add a new species to the Reaction class
	#Takes care of checking if it is already there or not, does nothing if it is
	#So maybe "ensureSpeciesIsRecored" would be a better name but that's a mouthful
	#Will assume an initial concentration of 0 if one is not given
	def addSpecies(self, spec, C0=0):
		if not (spec in self.species.keys()):
			self.species[spec] = len(self.species.keys())
			self.C0[spec] = C0
	
	#Returns the instantaneous rate given the concentrations indexed in an array
	def dCdt(self, C):
		toReturn = np.zeros(len(C))
		for rate in self.rates:
			toReturn += rate.calcRate(C)
		#print("dCdt:", ",".join(["{:.2e}"]*len(C)).format(*C), ",".join(["{:.2e}"]*len(toReturn)).format(*toReturn))
		return toReturn
	
	#Integrates the rates, by default from 0 to inf
	#Initial concentrations are assumed to be 0 unless given in the dict C0
	def integrate(self, t_init=None, t_eval=None, t_final=None):
		
		if t_init is not None:
			self.t_init = t_init
		if t_eval is not None:
			self.t_eval = t_eval
			#Note that None is a valid option for this
			#So it'll need to be manually set back to None if you actually want None
			#Keeping these inputs mostly for backwards compatibility, originally Rxn didn't keep track of time
			#So I'm not worrying about a little inconvenience here
		if t_final is not None:
			self.t_final = t_final

		#Get an array of initial concentrations
		C0_indexed = np.zeros(len(self.species.keys()))
		for specie in self.C0.keys():
			C0_indexed[self.species[specie]] = self.C0[specie]
		
		#Integrate, save solution as the solution for the current rates
		self.sol = solve_ivp(lambda t,C: self.dCdt(C), (self.t_init, self.t_final), C0_indexed, method='LSODA', t_eval=self.t_eval)
		
		return self.sol.y
	
	#Updates the ks for rates and integrates the new ks
	def iterate_rates(self, ks, t_init=None, t_eval=None, t_final=None):
		
		#Update the rates with the latest and greatest guesses
		mutables = self.mutableRates()
		for i in range(len(mutables)):
			mutables[i].setRate(ks[i])
		
		#Integrate to update the reaction's solution
		return self.integrate(t_init, t_eval, t_final)
	
	#Get the concentrations after integrating the reaction from C0 using the given rate laws
	#each species' concentration is given as a row in the returned array, columns being time points
	def getConcentration(self, species):
		toReturn = np.empty((len(species),self.sol.y.shape[1]))
		for i in range(len(species)):
			toReturn[i] = self.sol.y[self.species[species[i]]]
		return toReturn
	
	#Getting the gradient with respect to changes in kinetic parameters
	#fix_intrument determines if the instrument parameters are refit to new simulated C or not
	#For fitting I believe fixing is good; for errors you want to not fix it
	def gradient(self, instrument, visible_species=None, mutable_C0s=None, t_init_fixed=True, fix_instrument=False):
		
		if mutable_C0s is None:
			mutable_C0s = []
		
		#Here None actually means all
		if visible_species is None: visible_species = list(self.species.keys())
		
		mutable_ks = self.mutableRates()
		
		toReturn = np.zeros((len(mutable_ks)+len(mutable_C0s)+(0 if t_init_fixed else 1), instrument.size()))
		
		self.integrate()
		center = instrument.get_residuals(self.getConcentration(visible_species))
		
		k = 0
		for rate in mutable_ks:
			curr = rate.getRate()
			delta = 1e-9 if curr == 0 else curr*0.1 if curr < 1e-9 else curr*0.01
			rate.setRate(curr+delta)
			self.integrate()
			up = instrument.get_residuals(self.getConcentration(visible_species), fixed = fix_instrument)
			if curr - delta >= 0:
				rate.setRate(curr-delta)
				self.integrate()
				down = instrument.get_residuals(self.getConcentration(visible_species), fixed = fix_instrument)
				toReturn[k] = (up-down)/(2*delta)
			else:
				toReturn[k] = (up-center)/delta
			rate.setRate(curr)
			k += 1
		for mutable_C0 in mutable_C0s:
			curr = self.C0[mutable_C0]
			delta = 1e-9 if curr == 0 else curr*0.1 if curr < 1e-9 else curr*0.01
			self.C0[mutable_C0] = curr+delta
			self.integrate()
			up = instrument.get_residuals(self.getConcentration(visible_species), fixed = fix_instrument)
			if curr-delta >= 0:
				self.C0[mutable_C0] = curr-delta
				self.integrate()
				up = instrument.get_residuals(self.getConcentration(visible_species), fixed = fix_instrument)
				toReturn[k] = (up-down)/(2*delta)
				self.C0[mutable_C0] = curr
			else:
				toReturn[k] = (up-center)/delta
			k += 1
		if not t_init_fixed:
			curr = self.init_t
			delta = np.max(curr*0.01, 1)
			self.init_t = curr+delta
			self.integrate()
			up = instrument.residuals(self.getConcentration(visible_species), fixed = fix_instrument)
			self.init_t = curr-delta
			self.integrate()
			up = instrument.residuals(self.getConcentration(visible_species), fixed=fix_instrument)
			toReturn[k] = (up-down)/(2*delta)
			self.init_t = curr
			k += 1
		return toReturn.T
			
	#Fits the rate constants of this Reaction instance's Rate instances to experimental data
	#	C0 gives the initial concentrations - if None assumes C0 values have already been given
	#	instrument gives experimental data; is of a class defined below which is capable of
	#		figuring out how well simulated concentrations fits to experimental data
	#	visible_species gives which species are actually represented in instrument, assumed to
	#		be all data if None is given
	#	mutable_C0s gives species whose initial concentration is unknown and an attempt should
	#		be made to fit it (in general this probably risks underdetermining the system, so try
	#		to come up with good C0 values that don't need fit if possible)
	#	model_t_init indicates whether or not the inital time of the reaction needs to be fit
	#		i.e., you don't know the exact moment of injection relative to UV-vis scan times 
	#		I've had mixed success with it, coming up with a good estimate of at what
	#		time C(t) = C0 and adjusting instrument accordingly is probably best
	#Note that I changed this since first writing - there might be a script or two floating
	#	around which it isn't backwards compatible with
	def fitExperiment(self, C0, instrument, visible_species=None, mutable_C0s=None, model_t_init=False):
		
		#They need to give C0, but if it has been given otherwise they can just give None
		if C0 is not None:
			for species in self.C0:
				self.C0[species] = C0[species] if species in C0 else 0
		if mutable_C0s is None:
			mutable_C0s = []
		
		#Here None actually means all
		if visible_species is None: visible_species = list(self.species.keys())
		
		#Guessing ks means iterating and then having the instrument get residuals
		def guess_ks(ks):
			#Get the times to integrate over
			if model_t_init:
				t_init = ks[-1]
				t_toEval = instrument.get_times()
				ks = ks[:-1]
			else:
				t_init = 0
				t_toEval = instrument.get_times()
			for i in range(len(mutable_C0s)):
				self.C0[mutable_C0s[i]] = ks[i-len(mutable_C0s)]
			ks = ks[:len(ks)-len(mutable_C0s)]
			#Iterate the rates and get the residuals
			self.iterate_rates(ks, t_init, t_toEval, t_toEval[-1])
			return instrument.get_residuals(self.getConcentration(visible_species))
		
		def jacobian_ks(ks, fix_instrument):
			#Get the times to integrate over
			if model_t_init:
				self.t_init = ks[-1]
				ks = ks[:-1]
			self.t_eval = instrument.get_times()
			self.t_final = self.t_eval[-1]
			for i in range(len(mutable_C0s)):
				self.C0[mutable_C0s[i]] = ks[i-len(mutable_C0s)]
			ks = ks[:len(ks)-len(mutable_C0s)]
			i = 0
			for rate in self.mutableRates():
				rate.setRate(ks[i])
				i += 1
			return self.gradient(instrument, visible_species=visible_species, mutable_C0s=mutable_C0s, t_init_fixed=not model_t_init, fix_instrument=fix_instrument)
		
		#Get the current rates as initial guesses
		mutable_ks = self.mutableRates()
		k0s = np.zeros(len(mutable_ks) + len(mutable_C0s) + (1 if model_t_init else 0))
		j = 0
		for i in range(len(mutable_ks)):
			k0s[i] = mutable_ks[i].getRate()
		for i in range(len(mutable_C0s)):
			k0s[len(mutable_ks)+i] = self.C0[mutable_C0s[i]]
		
		#Now guess and check to find the k
		if not model_t_init:
			bounds = (0, np.inf)
		else:
			bounds = np.zeros(2, len(k0s))
			bounds[1,:] = np.inf
			bounds[0,-1] = -np.inf
		iters = 0
		while iters < 10: #10 is maximum number of iterations
			sol = least_squares(guess_ks, k0s, jac = lambda ks: jacobian_ks(ks, True), bounds=bounds)
			sol = least_squares(guess_ks, sol.x, jac = '3-point', bounds=bounds)
			k0s = sol.x
			if iters != 0 and sol.cost > 0.95*prev_cost:
				break
			prev_cost = sol.cost
			iters += 1
		if iters == 10:
			print("Warning: Reached Maximum Number of Iterations (this warning is not logged in any output file)")
		guess_ks(sol.x)
		self.k_fit = []
		for i in range(len(self.rates)):
			if self.rates[i].getMutable():
				self.k_fit.append(i)
		self.C0_fit = mutable_C0s
		self.t_fit = model_t_init
		self.t_init = sol.x[-1] if model_t_init else 0
		self.t_eval = instrument.get_times()
		self.t_final = instrument.get_times()[-1]
		self.cov = 2*sol.cost*np.linalg.inv(sol.jac.T @ sol.jac)/(len(sol.fun)-len(sol.x))
		
		#invStds = np.diag(1/np.sqrt(np.einsum('ii->i', self.cov)))
		#print("Scaled Covariances:")
		#print( invStds @ self.cov @ invStds )
		#print(sol)
		
		return sol
	
	#Plot simulated concentrations with the given pyplot axs
	def plotConcentrations(self, species, axs):
		for specie in species:
			axs.plot(self.sol.t, self.sol.y[self.species[specie]], '-', label=specie)
		axs.set_xlabel("Time (s)")
		axs.set_ylabel("Concentration")
	
	#Prints a report of an optimized Reaction with all the rate constants and errors, etc.
	#You can give it a file.write function to print to file instead of standard stream, but
	#	note that it assumes output has print's default ending of "\n" on it
	def report(self, output=print):
		output("\n\nReaction Scheme:\n")
		
		lhs, rhs, rev = [], [], []
		maxLengthL = 0
		for rate in self.rates:
			lhs.append(rate.reactantStr())
			rhs.append(rate.productStr())
			rev.append(isinstance(rate, Reversible))
			if len(lhs[-1]) > maxLengthL: maxLengthL = len(lhs[-1])
		template = "{{:>3}}. {{:>{}}} {{}}-> {{}}".format(maxLengthL)
		for i in range(len(lhs)):
			output(template.format(i, lhs[i], "<" if rev[i] else "-", rhs[i]).format(*self.species.keys()))
		
		output("\n\nInitial Concentrations:\n")
		for spec in self.C0:
			if self.C0_fit is not None and spec in self.C0_fit:
				i = 0
				while self.C0_fit[i] != spec: i += 1
				i += len(self.k_fit)
				output("\t[{}] = {:.2e}({:.2e})".format(spec, self.C0[spec], self.cov[i,i]**0.5))
			else:
				output("\t[{}] = {}".format(spec, self.C0[spec]))
	
		output("\n\nRate Laws with optimized experimental fit:\n")
		
		def format_rateLaw(k, std, rest):
			if std != 0:
				return "Rate = {:9.2e}({:9.2e}) {}".format(k, std, rest)
			else:
				return "Rate = {:9.2e} {}".format(k, rest)
		j = 0
		for i in range(len(self.rates)):
			if self.k_fit is not None and i in self.k_fit:
				cov = self.cov[j,j]
				j += 1
			else:
				cov = 0
			if isinstance(self.rates[i], Reversible):
				newLaws = self.rates[i].rateLaw()
				output(format_rateLaw(self.rates[i].getRate(), cov**0.5, newLaws[0]).format(*self.species.keys()))
				output(format_rateLaw(self.rates[i].getRate()/self.rates[i].K, cov**0.5/self.rates[i].K, newLaws[1]).format(*self.species.keys()))
			else:
				output(format_rateLaw(self.rates[i].getRate(), cov**0.5, self.rates[i].rateLaw()).format(*self.species.keys()))
		
		output("\n(Note: For reversible reactions, the equilibrium constant was fixed and the errors do not take into account uncertainty in this fixed value)\n")
		
		if self.cov is not None:
			output("Full Covariance Matrix:")
			for i in range(self.cov.shape[0]):
				nextLine = ""
				for j in range(self.cov.shape[1]):
					nextLine = nextLine + "{:.2e} ".format(self.cov[i,j])
				output(nextLine)
			output("")

#Represents experimental data to fit simulated Reaction concentrations to
class ConcentrationInstrument:
	def get_residuals(self, concentrations): pass
	def size(self): pass
	def get_simulated(self): pass
	def get_experimental(self): pass
	def get_times(self): pass

#Instrument based on spectral data "condensed" with SVD
class AbstractSignals(ConcentrationInstrument):
	
	def __init__(self, times, abstracts):
		self.times = times
		self.abstracts = abstracts
		self.converter = None
	
	def size(self):
		return self.abstracts.size
	
	#The abstract concentrations from SVD are assumed to be linear
	#	combinations of "true" concentrations:
	#	C_abstract = A * C_true
	#If fixed, the old converter A is used to generate simulated abstracts
	#If not fixed, the "best" converter A is found by pseudoinverse
	def get_residuals(self, concentrations, fixed=False):
		if not fixed:
			self.converter = self.abstracts @ np.linalg.pinv(concentrations)
		self.simulated = self.converter @ concentrations
		toReturn =  self.simulated - self.abstracts
		if len(toReturn.shape) == 1:
			return toReturn
		else:
			return np.concatenate(toReturn)
	
	def get_simulated(self):
		return self.simulated
	
	def get_experimental(self):
		return self.abstracts
	
	def get_times(self):
		return self.times