#Script to handle tunneling corrections
import numpy as np
from scipy.integrate import quad, solve_ivp
from scipy.optimize import root_scalar, least_squares
from scipy.stats import linregress

#By default, units are:
#	Length in Angstrom
#	Energy in kcal/mol
#	Mass in AMU
#	Time in s
#	Temp in K

hbar = 1.518e-14
k = 1.987e-3
pi = 3.1415926535898
E_conv = 4.184e26 #This is how many amu A^2 / s^2 are in 1 kcal/mol
c = 2.998e10 #This is in cm/s, not A/s

class Eckart:
	
	#Given a barrier height, the reaction energy, an the imaginary frequency
	#The mass is needed to get the force constant out of the frequency
	#Frequency is really expected in wavenumber, so its converted from 1/cm to 1/s
	def __init__(self, E_act, E_rct, v, mass = 1):
		self.E_act = E_act
		self.E_rct = E_rct
		self.v = v
		self.forcek = mass * (2*pi*v*c)**2
		self.A = E_rct
		self.B = (np.sqrt(E_act)+np.sqrt(E_act-E_rct))**2
		self.Lbar = np.sqrt(2*E_conv/self.forcek)/((1/np.sqrt(E_act))+(1/np.sqrt(E_act-E_rct)))
		
		self.x0 = self.Lbar*np.log((self.A+self.B)/(self.B-self.A)) #Bell p. 27
	
	#Calculated with Bell 2.45, but with a shift in x to center the maxima
	def potential(self, x):
		if x == np.inf:
			return self.A
		x += self.x0
		y = np.exp(x/self.Lbar)
		return self.A*y/(1+y) + self.B*y/(1+y)**2
		
	#The formulas used are from Johnson and Heicklen
	def permeability(self, E, mass = 1):
		nu = np.sqrt(self.forcek/mass)/(2*pi)
		alpha = E/(hbar*nu)
		beta = (E-self.E_rct)/(hbar*nu)
		alpha1 = self.E_act/(hbar*nu)
		alpha2 = (self.E_act-self.E_rct)/(hbar*nu)
		denom = (1/np.sqrt(alpha1)) + (1/np.sqrt(alpha2))
		
		a = 2*np.sqrt(alpha)/denom
		b = 2*np.sqrt(beta)/denom
		d = 2*np.sqrt(np.abs(alpha1*alpha2 - pi*pi/4))
		
		thing1 = np.cosh(a-b)
		thing2 = np.cosh(a+b)
		thing3 = np.cosh(d)
		
		if thing2 == np.inf:
			return 1
		if thing3 == np.inf:
			return 0
		else:
			return ((np.cosh(a+b)-np.cosh(a-b))/(np.cosh(a+b)+np.cosh(d)))
			
	#Does the integration of permeability over various energies
	#Numerically integrate permeability(E) as given in Bell's equation 3.4
	def transmission(self, T, mass = 1):
		integral, _ = quad(lambda E: self.permeability(E, mass)*np.exp(-(E-self.potential(0))/(k*T))/(k*T), max(0, self.potential(np.inf)), np.inf)
		return integral